@@ -21,6 +21,7 @@
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE QuantifiedConstraints #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE UndecidableInstances #-}
@@ -30,7 +31,7 @@ import Control.DeepSeq
 import Data.Data(Data)
 import qualified Data.DList as DL
 import Data.Kind (Type)
-import Data.List(foldl', zipWith4, zipWith5, sortBy, sortOn)
+import Data.List(foldl', zipWith4, zipWith5, sortBy, sortOn, foldl1')
 import Data.Proxy
 import GHC.Exts(Constraint, build)
 import GHC.Generics(Generic)
@@ -45,6 +46,8 @@ import Text.PrettyPrint.HughesPJClass
 -- constraint on the elements, which VecElem allows you to express.
 -- For vector types that don't need the constraint it can be set
 -- to some dummy class.
+-- | The 'Vector' class is the interface to the underlying storage for the arrays.
+-- The operations map straight to operations for 'Vector'.
 class Vector v where
   type VecElem v :: Type -> Constraint
   vIndex    :: (VecElem v a) => v a -> Int -> a
@@ -113,9 +116,6 @@ instance Vector [] where
 prettyShowL :: (Pretty a) => PrettyLevel -> a -> String
 prettyShowL l = render . pPrintPrec l 0
 
--- We expect all N to be non-negative, but we use Int for convenience.
-type N = Int
-
 -- | The type /T/ is the internal type of arrays.  In general,
 -- operations on /T/ do no sanity checking as that should be done
 -- at the point of call.
@@ -125,22 +125,27 @@ type N = Int
 -- dimension starts you calculate vector index @offset + i*strides[0]@.
 -- To find where item /i,j/ of the two outermost dimensions is you
 -- calculate vector index @offset + i*strides[0] + j*strides[1]@, etc.
+type role T representational nominal
 data T v a = T
-    { strides :: [N]      -- length is tensor rank
-    , offset  :: !N       -- offset into vector of values
+    { strides :: ![Int]   -- length is tensor rank
+    , offset  :: !Int     -- offset into vector of values
     , values  :: !(v a)   -- actual values
     }
     deriving (Show, Generic, Data)
 
 instance NFData (v a) => NFData (T v a)
 
+-- | The shape of an array is a list of its dimensions.
 type ShapeL = [Int]
 
 badShape :: ShapeL -> Bool
 badShape = any (< 0)
 
 -- When shapes match, we can be efficient and use loop-fused comparisons instead
--- of materializing a list.
+-- of materializing a vector.
+-- Note this assumes the shape is the same for both Vectors.
+-- TODO(augustss): if the array is a small fraction of the vector this can be inefficient.
+{-# INLINABLE equalT #-}
 equalT :: (Vector v, VecElem v a, Eq a, Eq (v a))
                   => ShapeL -> T v a -> T v a -> Bool
 equalT s x y | strides x == strides y
@@ -149,6 +154,7 @@ equalT s x y | strides x == strides y
              | otherwise = toVectorT s x == toVectorT s y
 
 -- Note this assumes the shape is the same for both Vectors.
+{-# INLINABLE compareT #-}
 compareT :: (Vector v, VecElem v a, Ord a, Ord (v a))
             => ShapeL -> T v a -> T v a -> Ordering
 compareT s x y = compare (toVectorT s x) (toVectorT s y)
@@ -156,7 +162,7 @@ compareT s x y = compare (toVectorT s x) (toVectorT s y)
 -- Given the dimensions, return the stride in the underlying vector
 -- for each dimension.  The first element of the list is the total length.
 {-# INLINE getStridesT #-}
-getStridesT :: ShapeL -> [N]
+getStridesT :: ShapeL -> [Int]
 getStridesT = scanr (*) 1
 
 -- Convert an array to a list by indexing through all the elements.
@@ -181,7 +187,7 @@ toListT sh a@(T ss0 o0 v)
 -- | Check if the strides are canonical, i.e., if the vector have the natural layout.
 -- XXX Copy special cases from Tensor.
 {-# INLINE isCanonicalT #-}
-isCanonicalT :: (Vector v, VecElem v a) => [N] -> T v a -> Bool
+isCanonicalT :: (Vector v, VecElem v a) => [Int] -> T v a -> Bool
 isCanonicalT (n:ss') (T ss o v) =
     o == 0 &&         -- Vector offset is 0
     ss == ss' &&      -- All strides are normal
@@ -234,13 +240,14 @@ toVectorT sh a@(T ats ao v) =
 
 -- Convert to a vector containing the right elements,
 -- but not necessarily in the right order.
+-- This is used for reduction with commutative&associative operations.
 {-# INLINE toUnorderedVectorT #-}
 toUnorderedVectorT :: (Vector v, VecElem v a) => ShapeL -> T v a -> v a
 toUnorderedVectorT sh a@(T ats ao v) =
   -- Figure out if the array maps onto some contiguous slice of the vector.
   -- Do this by checking if a transposition of the array corresponds to
   -- normal strides.
-  -- First sort the strides in descending order, amnd rearrange the shape the same way.
+  -- First sort the strides in descending order, and rearrange the shape the same way.
   -- Then compute the strides from this rearranged shape; these will be the normal
   -- strides for this shape.  If these strides agree with the sorted actual strides
   -- it is a transposition, and we can just slice out the relevant piece of the vector.
@@ -260,12 +267,12 @@ fromVectorT sh = T (tail $ getStridesT sh) 0
 
 -- Convert from a list
 {-# INLINE fromListT #-}
-fromListT :: (Vector v, VecElem v a) => [N] -> [a] -> T v a
+fromListT :: (Vector v, VecElem v a) => [Int] -> [a] -> T v a
 fromListT sh = fromVectorT sh . vFromList
 
 -- Index into the outermost dimension of an array.
 {-# INLINE indexT #-}
-indexT :: T v a -> N -> T v a
+indexT :: T v a -> Int -> T v a
 indexT (T (s : ss) o v) i = T ss (o + i * s) v
 indexT _ _ = error "impossible"
 
@@ -353,7 +360,7 @@ subArraysT sh ten = sub sh ten []
 
 -- Reverse the given dimensions.
 {-# INLINE reverseT #-}
-reverseT :: [N] -> ShapeL -> T v a -> T v a
+reverseT :: [Int] -> ShapeL -> T v a -> T v a
 reverseT rs sh (T ats ao v) = T rts ro v
   where (ro, rts) = rev 0 sh ats
         rev !_ [] [] = (ao, [])
@@ -390,45 +397,86 @@ allSameT sh t@(T _ _ v)
         !x = vIndex v' 0
     in  vAll (x ==) v'
 
+newtype Rect = Rect { unRect :: [String] }  -- A rectangle of text
+
+toRect :: String -> Rect
+toRect = Rect . lines
+
+fromRect :: Rect -> String
+fromRect (Rect ls) = unlines ls
+
+-- Make each Rect be of size h * w
+rectPad :: Int -> Int -> Rect -> Rect
+rectPad h w (Rect ls) = Rect $ map padL ls ++ replicate (h - length ls) mt
+  where mt = replicate w ' '
+        padL s = replicate (w - length s) ' ' ++ s
+
+-- Horizontal catenation.  Assumes input rectangle are padded.
+-- Adds empty space between Rects.
+hcatRect :: Rect -> Rect -> Rect
+hcatRect (Rect xs) (Rect ys) = Rect $ zipWith (\ x y -> x ++ " " ++ y) xs ys
+
+-- Vertical catenation.  Assumes input rectangle are padded.
+-- Adds no space between Rects.
+vcatRect :: Rect -> Rect -> Rect
+vcatRect (Rect xs) (Rect ys) = Rect $ xs ++ ys
+
+rectHeight :: Rect -> Int
+rectHeight = length . unRect
+
+-- Widest line
+rectWidth :: Rect -> Int
+rectWidth = maximum . (0:) . map length . unRect
+
 ppT
   :: (Vector v, VecElem v a, Pretty a)
   => PrettyLevel -> Rational -> ShapeL -> T v a -> Doc
-ppT l p sh = maybeParens (p > 10) . vcat . map text .  box prettyBoxMode . ppT_ (prettyShowL l) sh
+ppT l p sh = maybeParens (p > 10) . vcat' . map text . unRect . box boxMode . ppT_ (prettyShowL l) sh
+  where boxMode | l >= prettyNormal = BoxMode True True True
+                | otherwise = BoxMode False False False
+        vcat' = foldl' ($+$) empty
 
 ppT_
   :: (Vector v, VecElem v a)
-  => (a -> String) -> ShapeL -> T v a -> String
-ppT_ show_ sh t = revDropWhile (== '\n') $ showsT sh t' ""
-  where ss = map show_ $ toListT sh t
-        n = maximum $ map length ss
-        ss' = map padSP ss
-        padSP s = replicate (n - length s) ' ' ++ s
-        t' :: T [] String
+  => (a -> String) -> ShapeL -> T v a -> Rect
+ppT_ show_ sh t = showsT sh t'
+  where ss = map (toRect . show_) $ toListT sh t
+        maxH = maximum $ map rectHeight ss
+        maxW = maximum $ map rectWidth ss
+        ss' = map (rectPad maxH maxW) ss
+        t' :: T [] Rect
         t' = T (tail (getStridesT sh)) 0 ss'
 
-showsT :: [N] -> T [] String -> ShowS
-showsT (0:_)  _ = showString "EMPTY"
-showsT []     t = showString $ unScalarT t
-showsT s@[_]  t = showString $ unwords $ toListT s t
-showsT (n:ns) t =
-    foldr (.) id [ showsT ns (indexT t i) . showString "\n" | i <- [0..n-1] ]
+showsT :: [Int] -> T [] Rect -> Rect
+showsT []     t = unScalarT t
+showsT s@[_]  t = foldl1' hcatRect $ toListT s t
+showsT (n:ns) t = foldl1' vcat' rs
+  where vcat' x y = vcatRect x (vcatRect spc y)
+        spc = Rect $ replicate (length ns - 1) (replicate (rectWidth (head rs)) ' ')
+        rs = [ showsT ns (indexT t i) | i <- [0..n-1] ]
 
 data BoxMode = BoxMode { _bmBars, _bmUnicode, _bmHeader :: Bool }
 
 prettyBoxMode :: BoxMode
 prettyBoxMode = BoxMode False False False
 
-box :: BoxMode -> String -> [String]
-box BoxMode{..} s =
+-- Possibly draw a box around a (padded) rectangle.
+box :: BoxMode -> Rect -> Rect
+box BoxMode{..} (Rect ls) =
   let bar | _bmUnicode = '\x2502'
           | otherwise = '|'
-      ls = lines s
+      dash | _bmUnicode = '\x2500'
+           | otherwise = '-'
       ls' | _bmBars = map (\ l -> if null l then l else [bar] ++ l ++ [bar]) ls
           | otherwise = ls
-      h = "+" ++ replicate (length (head ls)) '-' ++ "+"
-      ls'' | _bmHeader = [h] ++ ls' ++ [h]
+      h = replicate (length (head ls)) dash
+      t | _bmUnicode = "\x250c" ++ h ++ "\x2510"
+        | otherwise = "+" ++ h ++ "+"
+      b | _bmUnicode = "\x2514" ++ h ++ "\x2518"
+        | otherwise = t
+      ls'' | _bmHeader = [t] ++ ls' ++ [b]
            | otherwise = ls'
-  in  ls''
+  in  Rect ls''
 
 zipWithLong2 :: (a -> b -> b) -> [a] -> [b] -> [b]
 zipWithLong2 f (a:as) (b:bs) = f a b : zipWithLong2 f as bs
@@ -448,7 +496,7 @@ padT v aps ash at = (ss, fromVectorT ss $ vConcat $ pad' aps ash st at)
 -- size 1, in which case it can be done by just manipulating
 -- the strides.  Given the old strides, the old shapes, and the
 -- new shape it will return the possible new strides.
-simpleReshape :: [N] -> ShapeL -> ShapeL -> Maybe [N]
+simpleReshape :: [Int] -> ShapeL -> ShapeL -> Maybe [Int]
 simpleReshape osts os ns
   | filter (1 /=) os == filter (1 /=) ns = Just $ loop ns sts'
     -- Old and new dimensions agree where they are not 1.
@@ -462,18 +510,22 @@ simpleReshape osts os ns
       loop _ _ = error $ "simpleReshape: shouldn't happen: " ++ show (osts, os, ns)
 simpleReshape _ _ _ = Nothing
 
+-- Note: assumes + is commutative&associative.
 {-# INLINE sumT #-}
 sumT :: (Vector v, VecElem v a, Num a) => ShapeL -> T v a -> a
 sumT sh = vSum . toUnorderedVectorT sh
 
+-- Note: assumes * is commutative&associative.
 {-# INLINE productT #-}
 productT :: (Vector v, VecElem v a, Num a) => ShapeL -> T v a -> a
 productT sh = vProduct . toUnorderedVectorT sh
 
+-- Note: assumes max is commutative&associative.
 {-# INLINE maximumT #-}
 maximumT :: (Vector v, VecElem v a, Ord a) => ShapeL -> T v a -> a
 maximumT sh = vMaximum . toUnorderedVectorT sh
 
+-- Note: assumes min is commutative&associative.
 {-# INLINE minimumT #-}
 minimumT :: (Vector v, VecElem v a, Ord a) => ShapeL -> T v a -> a
 minimumT sh = vMinimum . toUnorderedVectorT sh
