@@ -17,6 +17,7 @@
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE UndecidableInstances #-}
 -- | Arrays of dynamic size.  The arrays are polymorphic in the underlying
@@ -54,6 +55,7 @@ import Text.PrettyPrint.HughesPJClass hiding ((<>))
 import Data.Array.Internal
 
 -- | Arrays stored in a /v/ with values of type /a/.
+type role Array representational nominal
 data Array v a = A !ShapeL !(T v a)
   deriving (Generic, Data)
 
@@ -92,7 +94,7 @@ size = product . shapeL
 shapeL :: Array v a -> ShapeL
 shapeL (A s _) = s
 
--- | The rank of an array, i.e., the number if dimensions it has.
+-- | The rank of an array, i.e., the number of dimensions it has.
 -- O(1) time.
 {-# INLINE rank #-}
 rank :: Array v a -> Int
@@ -352,8 +354,8 @@ rotate d k a = error $ "Incorrect arguments to rotate: " ++ show (d, k, rank a)
 
 -- | Extract a slice of an array.
 -- The first argument is a list of (offset, length) pairs.
--- The length of the slicing argument must not exceed the rank of the arrar.
--- The extracted slice mul fall within the array dimensions.
+-- The length of the slicing argument must not exceed the rank of the array.
+-- The extracted slice must fall within the array dimensions.
 -- E.g. @slice [1,2] (fromList [4] [1,2,3,4]) == [2,3]@.
 -- O(1) time.
 {-# INLINE slice #-}
@@ -473,7 +475,7 @@ anyA p (A sh t) = anyT sh p t
 -- | Test if the predicate holds for all elements.
 {-# INLINE allA #-}
 allA :: (Vector v, VecElem v a) => (a -> Bool) -> Array v a -> Bool
-allA p (A sh t) = anyT sh p t
+allA p (A sh t) = allT sh p t
 
 -- | Put the dimensions of the argument into the specified dimensions,
 -- and just replicate the data along all other dimensions.
