@@ -23,6 +23,7 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications #-}
 {-# LANGUAGE TypeFamilies #-}
@@ -63,10 +64,12 @@ import Text.PrettyPrint.HughesPJClass hiding ((<>))
 import Data.Array.Internal
 
 -- | Arrays stored in a /v/ with values of type /a/.
-data Array (n :: Nat) v a = A ShapeL (T v a)
+type role Array nominal representational nominal
+data Array (n :: Nat) v a = A !ShapeL !(T v a)
   deriving (Generic, Data)
 
 instance (Vector v, Show a, VecElem v a) => Show (Array n v a) where
+  {-# INLINABLE showsPrec #-}
   showsPrec p a@(A s _) = showParen (p > 10) $
     showString "fromList " . showsPrec 11 s . showString " " . showsPrec 11 (toList a)
 
@@ -91,6 +94,7 @@ instance (NFData (v a)) => NFData (Array n v a) where
   rnf (A sh v) = rnf sh `seq` rnf v
 
 -- | The number of elements in the array.
+-- O(1) time.
 {-# INLINE size #-}
 size :: Array n v a -> Int
 size = product . shapeL
@@ -103,7 +107,7 @@ size = product . shapeL
 shapeL :: Array n v a -> ShapeL
 shapeL (A s _) = s
 
--- | The rank of an array, i.e., the number if dimensions it has.
+-- | The rank of an array, i.e., the number of dimensions it has.
 -- O(1) time.
 {-# INLINE rank #-}
 rank :: forall n v a . (KnownNat n) => Array n v a -> Int
@@ -118,7 +122,7 @@ index (A (s:ss) t) i | i < 0 || i >= s = error $ "index: out of bounds " ++ show
 index (A [] _) _ = error "index: scalar"
 
 -- | Convert to a list with the elements in the linearization order.
--- O(1) time.
+-- O(n) time.
 {-# INLINE toList #-}
 toList :: (Vector v, VecElem v a) => Array n v a -> [a]
 toList (A sh t) = toListT sh t
@@ -168,7 +172,7 @@ reshape :: forall n n' v a . (HasCallStack,Vector v, VecElem v a, KnownNat n, Kn
            ShapeL -> Array n v a -> Array n' v a
 reshape sh (A sh' t@(T ost oo v))
   | n /= n' = error $ "reshape: size mismatch " ++ show (sh, sh')
-  | length sh /= valueOf @n' = error $ "reshape: rank mismatch " ++ show (length sh, valueOf @n :: Int)
+  | length sh /= valueOf @n' = error $ "reshape: rank mismatch " ++ show (length sh, valueOf @n' :: Int)
   | vLength v == 1 = A sh $ T (map (const 0) sh) 0 v  -- Fast special case for singleton vector
   | Just nst <- simpleReshape ost sh' sh = A sh $ T nst oo v
   | otherwise = A sh $ T st 0 $ toVectorT sh' t
@@ -318,7 +322,7 @@ unravel = rerank @1 scalar
 {-# INLINE window #-}
 window :: forall n n' v a . (Vector v, KnownNat n, KnownNat n') =>
           [Int] -> Array n v a -> Array n' v a
-window aws _ | valueOf @n' /= length aws + valueOf @n = error $ "window: rank mismatch: " ++ show (valueOf @n :: Int, length aws, valueOf @n' :: Int)
+window aws _ | valueOf @n' /= length aws + valueOf @n = error $ "window: rank mismatch: " ++ show (valueOf @n' :: Int, length aws, valueOf @n :: Int)
 window aws (A ash (T ss o v)) = A (win aws ash) (T (ss' ++ ss) o v)
   where ss' = zipWith const ss aws
         win (w:ws) (s:sh) | w <= s = s - w + 1 : win ws sh
@@ -367,8 +371,8 @@ rotate k a = rerank @d @p @(p + 1) f a
 
 -- | Extract a slice of an array.
 -- The first argument is a list of (offset, length) pairs.
--- The length of the slicing argument must not exceed the rank of the arrar.
--- The extracted slice mul fall within the array dimensions.
+-- The length of the slicing argument must not exceed the rank of the array.
+-- The extracted slice must fall within the array dimensions.
 -- E.g. @slice [1,2] (fromList [4] [1,2,3,4]) == [2,3]@.
 -- O(1) time.
 {-# INLINE slice #-}
@@ -395,6 +399,7 @@ rerank f (A sh t) =
   subArraysT osh t
   where (osh, ish) = splitAt (valueOf @n) sh
 
+{-# INLINABLE ravelOuter #-}
 ravelOuter :: (Vector v, VecElem v a, KnownNat m) => ShapeL -> [Array n v a] -> Array m v a
 ravelOuter _ [] = error "ravelOuter: empty list"
 ravelOuter osh as | not $ allSame shs = error $ "ravelOuter: non-conforming inner dimensions: " ++ show shs
@@ -489,7 +494,7 @@ anyA p (A sh t) = anyT sh p t
 -- | Test if the predicate holds for all elements.
 {-# INLINE allA #-}
 allA :: (Vector v, VecElem v a) => (a -> Bool) -> Array r v a -> Bool
-allA p (A sh t) = anyT sh p t
+allA p (A sh t) = allT sh p t
 
 -- | Put the dimensions of the argument into the specified dimensions,
 -- and just replicate the data along all other dimensions.
