@@ -22,6 +22,7 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE RoleAnnotations #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications #-}
 {-# LANGUAGE TypeFamilies #-}
@@ -64,10 +65,12 @@ import Data.Array.Internal
 import Data.Array.Internal.Shape
 
 -- | Arrays stored in a /v/ with values of type /a/.
+type role Array nominal representational nominal
 newtype Array (sh :: [Nat]) v a = A (T v a)
   deriving (Generic, Data)
 
 instance (Vector v, Show a, VecElem v a, Shape sh, Show (v a)) => Show (Array sh v a) where
+  {-# INLINABLE showsPrec #-}
   showsPrec p a@(A _) = showParen (p > 10) $
     showString "fromList @" . showsPrec 11 (shapeL a) . showString" " . showsPrec 11 (toList a)
 
@@ -107,7 +110,7 @@ size _ = sizeP (Proxy :: Proxy sh)
 shapeL :: forall sh v a . (Shape sh) => Array sh v a -> ShapeL
 shapeL _ = shapeP (Proxy :: Proxy sh)
 
--- | The rank of an array, i.e., the number if dimensions it has.
+-- | The rank of an array, i.e., the number of dimensions it has.
 -- O(1) time.
 {-# INLINE rank #-}
 rank :: forall sh v a . (Shape sh, KnownNat (Rank sh)) => Array sh v a -> Int
@@ -123,7 +126,7 @@ index (A t) i | i < 0 || i >= s = error $ "index: out of bounds " ++ show i ++ "
   where s = valueOf @s
 
 -- | Convert to a list with the elements in the linearization order.
--- O(1) time.
+-- O(n) time.
 {-# INLINE toList #-}
 toList :: (Vector v, VecElem v a, Shape sh) => Array sh v a -> [a]
 toList a@(A t) = toListT (shapeL a) t
@@ -172,14 +175,14 @@ normalize = fromVector . toVector
 reshape :: forall sh' sh v a .
            (Vector v, VecElem v a, Shape sh, Shape sh', Size sh ~ Size sh') =>
            Array sh v a -> Array sh' v a
-reshape a = reshape' (shapeL a) (shapeP (Proxy :: Proxy sh')) a
+reshape a = reshape' (shapeP (Proxy :: Proxy sh')) (shapeL a) a
 
 reshape' :: (Vector v, VecElem v a) =>
             ShapeL -> ShapeL -> Array sh v a -> Array sh' v a
 reshape' sh sh' (A t@(T ost oo v))
   | vLength v == 1 = A $ T (map (const 0) sh) 0 v  -- Fast special case for singleton vector
-  | Just nst <- simpleReshape ost sh sh' = A $ T nst oo v
-  | otherwise = A $ fromVectorT sh' $ toVectorT sh t
+  | Just nst <- simpleReshape ost sh' sh = A $ T nst oo v
+  | otherwise = A $ fromVectorT sh $ toVectorT sh' t
 
 -- | Change the size of dimensions with size 1.  These dimension can be changed to any size.
 -- All other dimensions must remain the same.
@@ -430,7 +433,7 @@ anyA p a@(A t) = anyT (shapeL a) p t
 -- | Test if the predicate holds for all elements.
 {-# INLINE allA #-}
 allA :: (Vector v, VecElem v a, Shape sh) => (a -> Bool) -> Array sh v a -> Bool
-allA p a@(A t) = anyT (shapeL a) p t
+allA p a@(A t) = allT (shapeL a) p t
 
 -- | Put the dimensions of the argument into the specified dimensions,
 -- and just replicate the data along all other dimensions.
